#ifndef PERL_VALUE_H
#define PERL_VALUE_H

// my stuff
#include "PerlExceptions.H"

// C/C++
#include <string>
#ifdef HPUX_ACC
#include <iostream/iostream.h>
typedef int streamsize;
#else
#include <iostream.h>
#endif

// STL
#include <vector>

using std::string;
using std::pair;

// Duplicates of Perl definitions.  If Perl changes, this may have to change
// as well.  We don't include the Perl headers because we get a lot of
// extra junk that can cause other problems.
struct sv;
struct av;
struct hv;
struct he;
typedef struct sv SV;
typedef struct av AV;
typedef struct hv HV;
typedef struct he HE;

// A class that is a nice wrapper around a Perl SV*
class PerlValue {
public:
    class ptr;            // return value for array_iter.operator->()
    class hash_value_ptr; // return value for hash_iter.operator->()
    class array_iter;     // iterator over a Perl array
    class hash_iter;      // iterator over a Perl hash

    friend class array_iter;
    friend class hash_iter;

private:
    SV* value;                    // The actual value.
    unsigned int exception_flags; // Flags that tell us whether to throw an
    				  // exception when things go wrong

public:
    PerlValue(SV* _value, unsigned int _exception_flags=0) throw()
	: value(_value), exception_flags(_exception_flags) {}

    // Functions to get info about the object
    bool isDefined()         const throw();
    bool isInt()             const throw();
    bool isDouble()          const throw();
    bool isNumber()          const throw();
    bool isString()          const throw();
    bool isScalarRef()       const throw();
    bool isArrayRef()        const throw();
    bool isHashRef()         const throw();
    int size()              const throw(PerlBadCast);
    const char* typeStr()    const throw();

    // Conversion operators to automatically convert the object to the
    // proper type.  The types have to match, or we error out.
    int Int()                const throw(PerlBadCast);
    double Double()          const throw(PerlBadCast);
    string String()          const throw(PerlBadCast);
    bool Bool()              const throw();
    operator int()           const throw(PerlBadCast) { return Int(); }
    operator double()        const throw(PerlBadCast) { return Double(); }
    operator string()        const throw(PerlBadCast) { return String(); }
    operator bool()          const throw() { return Bool(); }
    
    // Dereference a scalar reference
    PerlValue operator*()    const throw(PerlBadCast);
    ptr operator->()         const throw(PerlBadCast);

    // Access arrays and hashes
    PerlValue at(int array_index) const throw(PerlBadCast, PerlOutOfBounds);
    PerlValue at(const char* hash_key, size_t hash_key_length) const
	    throw(PerlBadCast, PerlOutOfBounds);
    PerlValue at(const char* hash_key) const
	    throw(PerlBadCast, PerlOutOfBounds);
    PerlValue at(const string& hash_key) const
	    throw(PerlBadCast, PerlOutOfBounds);
    PerlValue operator[](int array_index) const
	    throw(PerlBadCast, PerlOutOfBounds);
    PerlValue operator[](const char* hash_key) const
	    throw(PerlBadCast, PerlOutOfBounds);
    PerlValue operator[](const string& hash_key) const
	    throw(PerlBadCast, PerlOutOfBounds);

    // Functions for iterating over array and hash values
    array_iter array_begin() const throw(PerlBadCast);
    array_iter array_end()   const throw();
    hash_iter hash_begin()   const throw(PerlBadCast);
    hash_iter hash_end()     const throw();

    friend ostream& operator<<(ostream&, PerlValue) throw();
};
    
// A class to allow us to return a object from array_iter::operator->()
// that acts like a pointer to a PerlValue but doesn't have the lifetime
// problems of a pointer.
class PerlValue::ptr {
private:
    PerlValue value;

public:
    ptr(const PerlValue& _value) : value(_value) {}
    PerlValue* operator->() { return &value; }

private:
    ptr();               // not allowed
};

// A class to allow us to return a object from hash_iter::operator->()
// that acts like a pointer to a PerlValue but doesn't have the lifetime
// problems of a pointer.
class PerlValue::hash_value_ptr {
private:
    pair<string,PerlValue> value;

public:
    hash_value_ptr(const pair<string,PerlValue>& _value) : value(_value) {}
    pair<string,PerlValue>* operator->() { return &value; }

private:
    hash_value_ptr();    // not allowed
};

// A class patterned after the STL random access iterator to iterate over
// a Perl array.
class PerlValue::array_iter {
private:
    AV* array;           // The Perl array
    int index;           // The array item that this iterator is pointing to
    int array_size;      // The size of the Perl array
    unsigned int exception_flags;  // Exception flags to give to PerlValues

public:
    array_iter() throw()
	: array(0), index(0), array_size(0), exception_flags(0) {}
    array_iter(const PerlValue& array_ref) throw(PerlBadCast);

    // Get values from the iterator
    PerlValue operator*() const throw(PerlOutOfBounds);
    PerlValue::ptr operator->() const throw(PerlOutOfBounds);

    // Move the iterator forward or backward
    const array_iter& operator++() throw() { ++index; return *this; }
    array_iter operator++(int) throw(); // post increment
    const array_iter& operator--() throw() { --index; return *this; }
    array_iter operator--(int) throw(); // post decrement
    const array_iter& operator+=(int offset) throw();
    const array_iter& operator-=(int offset) throw();
    array_iter operator+(int offset) const throw();
    array_iter operator-(int offset) const throw();

    // Comparisons
    bool operator==(const array_iter& x) const throw();
    bool operator!=(const array_iter& x) const throw();

    friend array_iter operator+(int, const array_iter&) throw();

private:
    // default copy constructor, assignment operator, and destructor OK
};

// A class patterned after the STL random access iterator to iterate over
// a Perl array.  Due to limitations in the Perl API, the iterator cannot move
// backward and there cannot be more than one iterator per hash.
class PerlValue::hash_iter {
private:
    HV *hash;
    HE *hash_entry;
    unsigned int exception_flags;

public:
    hash_iter() throw() : hash(0), hash_entry(0), exception_flags(0) {}
    hash_iter(const PerlValue& hash_ref) throw(PerlBadCast);

    // Get values from the iterator
    pair<string,PerlValue> operator*() const throw(PerlOutOfBounds);
    hash_value_ptr operator->() const throw(PerlOutOfBounds);

    // Move the iterator forward
    hash_iter& operator++() throw();

    // Comparisons
    bool operator==(const hash_iter& x) const throw();
    bool operator!=(const hash_iter& x) const throw();

private:
    // default copy constructor, assignment operator, and destructor OK
};

inline PerlValue::ptr
PerlValue::operator->() const
throw(PerlBadCast)
{ return ptr(operator*()); }

inline PerlValue
PerlValue::at(const char* hash_key) const
throw(PerlBadCast, PerlOutOfBounds)
{ return at( hash_key, strlen(hash_key) ); }

// We use c_str() instead of data() because if length()
// is 0, Perl will try to figure out the length.
inline PerlValue
PerlValue::at(const string& hash_key) const
throw(PerlBadCast, PerlOutOfBounds)
{ return at( hash_key.c_str(), hash_key.length() ); }

inline PerlValue
PerlValue::operator[](int array_index) const
throw(PerlBadCast, PerlOutOfBounds)
{ return at(array_index); }

inline PerlValue
PerlValue::operator[](const char* hash_key) const
throw(PerlBadCast, PerlOutOfBounds)
{ return at(hash_key); }

inline PerlValue
PerlValue::operator[](const string& hash_key) const
throw(PerlBadCast, PerlOutOfBounds)
{ return at(hash_key); }

inline PerlValue::array_iter
PerlValue::array_begin() const
throw(PerlBadCast)
{ return array_iter(*this); }

inline PerlValue::array_iter
PerlValue::array_end() const
throw()
{ return array_iter(); }

inline PerlValue::hash_iter
PerlValue::hash_begin() const
throw(PerlBadCast)
{ return hash_iter(*this); }

inline PerlValue::hash_iter
PerlValue::hash_end() const
throw()
{ return hash_iter(); }

inline PerlValue::ptr
PerlValue::array_iter::operator->() const
throw(PerlOutOfBounds)
{ return ptr(operator*()); }

inline PerlValue::array_iter
PerlValue::array_iter::operator++(int) // post increment
throw()
{
    array_iter next = *this;
    ++index;
    return next;
}

inline PerlValue::array_iter
PerlValue::array_iter::operator--(int) // post decrement
throw()
{
    array_iter next = *this;
    --index;
    return next;
}

// We need to make comparisons against the end() value work
// in addition to regular comparisons.  The end() has an
// 'array' value of 0, so we return true if one is array_end()
// and the other has reached the end
inline bool
PerlValue::array_iter::operator==(const array_iter& x) const
throw()
{
    return (array == x.array && index == x.index) ||
	    (!array && x.index == x.array_size) ||
	    (!x.array && index == array_size);
}

inline bool
PerlValue::array_iter::operator!=(const array_iter& x) const
throw()
{ return !(*this == x); }

inline const PerlValue::array_iter&
PerlValue::array_iter::operator+=(int offset)
throw()
{ index += offset; return *this; }

inline const PerlValue::array_iter&
PerlValue::array_iter::operator-=(int offset)
throw()
{ index -= offset; return *this; }

inline PerlValue::array_iter
PerlValue::array_iter::operator+(int offset) const
throw()
{
    array_iter new_iter(*this);
    new_iter.index += offset;
    return new_iter;
}

inline PerlValue::array_iter
PerlValue::array_iter::operator-(int offset) const
throw()
{
    array_iter new_iter(*this);
    new_iter.index -= offset;
    return new_iter;
}

inline PerlValue::hash_value_ptr
PerlValue::hash_iter::operator->() const
throw(PerlOutOfBounds)
{
    return hash_value_ptr(operator*());
}

inline bool
PerlValue::hash_iter::operator==(const hash_iter& x) const
throw()
{ return hash_entry == x.hash_entry; }

inline bool
PerlValue::hash_iter::operator!=(const hash_iter& x) const
throw()
{ return hash_entry != x.hash_entry; }

inline PerlValue::array_iter
operator+(int offset, const PerlValue::array_iter& x)
throw()
{
    PerlValue::array_iter new_iter(x);
    new_iter.index += offset;
    return new_iter;
}

#endif
